==> polymer-virtual-entity-0.9.19+1.21.1-sources.jar: eu/pb4/polymer/virtualentity/api/elements/GenericEntityElement.java <==
package eu.pb4.polymer.virtualentity.api.elements;

import I;
import eu.pb4.polymer.virtualentity.api.VirtualEntityUtils;
import eu.pb4.polymer.virtualentity.api.tracker.DataTrackerLike;
import eu.pb4.polymer.virtualentity.api.tracker.EntityTrackedData;
import eu.pb4.polymer.virtualentity.api.tracker.SimpleDataTracker;
import it.unimi.dsi.fastutil.ints.IntList;
import org.jetbrains.annotations.Nullable;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.UUID;
import java.util.function.Consumer;
import net.minecraft.class_1297;
import net.minecraft.class_1299;
import net.minecraft.class_243;
import net.minecraft.class_2561;
import net.minecraft.class_2596;
import net.minecraft.class_2602;
import net.minecraft.class_2604;
import net.minecraft.class_2684;
import net.minecraft.class_2739;
import net.minecraft.class_3222;
import net.minecraft.class_3532;
import net.minecraft.class_4050;

@SuppressWarnings("ConstantConditions")
public abstract class GenericEntityElement extends AbstractElement {
    protected final DataTrackerLike dataTracker = this.createDataTracker();
    private final int id = VirtualEntityUtils.requestEntityId();
    private final UUID uuid = UUID.randomUUID();
    private float pitch;
    private float yaw;
    private boolean isRotationDirty;
    private boolean sendPositionUpdates = true;
    protected DataTrackerLike createDataTracker() {
        return new SimpleDataTracker(this.getEntityType());
    }

    public boolean isDirty() {
        return this.isRotationDirty || this.dataTracker.isDirty();
    }

    public boolean isRotationDirty() {
        return isRotationDirty;
    }

    public void ignorePositionUpdates() {
        setSendPositionUpdates(false);
    }

    public void setSendPositionUpdates(boolean b) {
        this.sendPositionUpdates = b;
    }

    public boolean isSendingPositionUpdates() {
        return this.sendPositionUpdates;
    }

    public void setPitch(float pitch) {
        if (this.pitch != pitch) {
            this.pitch = pitch;
            this.isRotationDirty = true;
        }
    }

    public void setYaw(float yaw) {
        if (this.yaw != yaw) {
            this.yaw = yaw;
            this.isRotationDirty = true;
        }
    }

    public float getYaw() {
        return this.yaw;
    }

    public float getPitch() {
        return this.pitch;
    }

    @Override
    public IntList getEntityIds() {
        return IntList.of(id);
    }

    public final UUID getUuid() {
        return this.uuid;
    }

    public final int getEntityId() {
        return this.id;
    }

    protected abstract class_1299<? extends class_1297> getEntityType();

    @Override
    public void startWatching(class_3222 player, Consumer<class_2596<class_2602>> packetConsumer) {
        packetConsumer.accept(this.createSpawnPacket(player));

        this.sendChangedTrackerEntries(player, packetConsumer);
    }

    protected class_2596<class_2602> createSpawnPacket(class_3222 player) {
        if (this.lastSyncedPos == null) {
            this.lastSyncedPos = this.getCurrentPos();
        }
        return new class_2604(this.id, this.uuid, this.lastSyncedPos.field_1352, this.lastSyncedPos.field_1351, this.lastSyncedPos.field_1350, this.pitch, this.yaw, this.getEntityType(), 0, class_243.field_1353, this.yaw);
    }

    protected void sendChangedTrackerEntries(class_3222 player, Consumer<class_2596<class_2602>> packetConsumer) {
        var changed = this.dataTracker.getChangedEntries();

        if (changed != null) {
            packetConsumer.accept(new class_2739(this.id, changed));
        }
    }

    @Override
    public void notifyMove(class_243 oldPos, class_243 newPos, class_243 delta) {
        if (this.sendPositionUpdates) {
            this.sendPositionUpdates();
        }
    }

    @Override
    public void stopWatching(class_3222 player, Consumer<class_2596<class_2602>> packetConsumer) {
    }

    @Override
    public void tick() {
        this.sendTrackerUpdates();
        if (this.sendPositionUpdates) {
            this.sendPositionUpdates();
        }
        this.sendRotationUpdates();
    }

    protected void sendPositionUpdates() {
        if (this.getHolder() == null) {
            return;
        }
        class_2596<class_2602> packet = null;
        var pos = this.getCurrentPos();

        if (pos.equals(this.lastSyncedPos)) {
            return;
        }

        if (this.lastSyncedPos == null) {
            var i = class_3532.method_15375(yaw * 256.0F / 360.0F);
            var j = class_3532.method_15375(pitch * 256.0F / 360.0F);
            packet = VirtualEntityUtils.createSimpleMovePacket(this.id, pos, (byte) i, (byte) j);
        } else {
            packet = VirtualEntityUtils.createMovePacket(this.id, this.lastSyncedPos, pos, this.isRotationDirty, this.yaw, this.pitch);
        }

        if (packet != null) {
            this.getHolder().sendPacket(packet);
            if (!(packet instanceof class_2684.class_2687)) {
                this.lastSyncedPos = pos;
            }
        }
        this.isRotationDirty = false;
    }

    protected void sendTrackerUpdates() {
        if (this.dataTracker.isDirty()) {
            var dirty = this.dataTracker.getDirtyEntries();
            if (dirty != null) {
                this.getHolder().sendPacket(new class_2739(this.id, dirty));
            }
        }
    }

    protected void sendRotationUpdates() {
        if (this.isRotationDirty) {
            var i = class_3532.method_15375(yaw * 256.0F / 360.0F);
            var j = class_3532.method_15375(pitch * 256.0F / 360.0F);
            this.getHolder().sendPacket(new class_2684.class_2687(id, (byte) i, (byte) j, false));
            this.isRotationDirty = false;
        }
    }

    public DataTrackerLike getDataTracker() {
        return this.dataTracker;
    }

    public class_4050 getPose() {
        return this.dataTracker.get(EntityTrackedData.POSE);
    }

    public void setPose(class_4050 pose) {
        this.dataTracker.set(EntityTrackedData.POSE, pose);
    }

    public void setOnFire(boolean onFire) {
        this.setFlag(EntityTrackedData.ON_FIRE_FLAG_INDEX, onFire);
    }

    protected boolean getFlag(int index) {
        return (this.dataTracker.get(EntityTrackedData.FLAGS) & 1 << index) != 0;
    }

    protected void setFlag(int index, boolean value) {
        byte b = this.dataTracker.get(EntityTrackedData.FLAGS);
        if (value) {
            this.dataTracker.set(EntityTrackedData.FLAGS, (byte) (b | 1 << index));
        } else {
            this.dataTracker.set(EntityTrackedData.FLAGS, (byte) (b & ~(1 << index)));
        }

    }

    public boolean isSneaking() {
        return this.getFlag(EntityTrackedData.SNEAKING_FLAG_INDEX);
    }

    public void setSneaking(boolean sneaking) {
        this.setFlag(EntityTrackedData.SNEAKING_FLAG_INDEX, sneaking);
    }

    public boolean isSprinting() {
        return this.getFlag(EntityTrackedData.SPRINTING_FLAG_INDEX);
    }

    public void setSprinting(boolean sprinting) {
        this.setFlag(EntityTrackedData.SPRINTING_FLAG_INDEX, sprinting);
    }

    public boolean isGlowing() {
        return this.getFlag(EntityTrackedData.GLOWING_FLAG_INDEX);
    }

    public final void setGlowing(boolean glowing) {
        this.setFlag(EntityTrackedData.GLOWING_FLAG_INDEX, glowing);
    }

    public boolean isInvisible() {
        return this.getFlag(EntityTrackedData.INVISIBLE_FLAG_INDEX);
    }

    public void setInvisible(boolean invisible) {
        this.setFlag(EntityTrackedData.INVISIBLE_FLAG_INDEX, invisible);
    }

    public int getAir() {
        return this.dataTracker.get(EntityTrackedData.AIR);
    }

    public void setAir(int air) {
        this.dataTracker.set(EntityTrackedData.AIR, air);
    }

    public int getFrozenTicks() {
        return this.dataTracker.get(EntityTrackedData.FROZEN_TICKS);
    }

    public void setFrozenTicks(int frozenTicks) {
        this.dataTracker.set(EntityTrackedData.FROZEN_TICKS, frozenTicks);
    }

    @Nullable
    public class_2561 getCustomName() {
        return this.dataTracker.get(EntityTrackedData.CUSTOM_NAME).orElse(null);
    }

    public void setCustomName(@Nullable class_2561 name) {
        this.dataTracker.set(EntityTrackedData.CUSTOM_NAME, Optional.ofNullable(name));
    }

    public boolean isCustomNameVisible() {
        return Boolean.TRUE == this.dataTracker.get(EntityTrackedData.NAME_VISIBLE);
    }

    public void setCustomNameVisible(boolean visible) {
        this.dataTracker.set(EntityTrackedData.NAME_VISIBLE, visible);
    }

    public boolean isSilent() {
        return Boolean.TRUE == this.dataTracker.get(EntityTrackedData.SILENT);
    }

    public void setSilent(boolean silent) {
        this.dataTracker.set(EntityTrackedData.SILENT, silent);
    }

    public boolean hasNoGravity() {
        return Boolean.TRUE == this.dataTracker.get(EntityTrackedData.NO_GRAVITY);
    }

    public void setNoGravity(boolean noGravity) {
        this.dataTracker.set(EntityTrackedData.NO_GRAVITY, noGravity);
    }
}

==> polymer-virtual-entity-0.14.0-rc.2+1.21.9-rc1-sources.jar: eu/pb4/polymer/virtualentity/api/elements/GenericEntityElement.java <==
package eu.pb4.polymer.virtualentity.api.elements;

import eu.pb4.polymer.virtualentity.api.VirtualEntityUtils;
import eu.pb4.polymer.virtualentity.api.tracker.DataTrackerLike;
import eu.pb4.polymer.virtualentity.api.tracker.EntityTrackedData;
import eu.pb4.polymer.virtualentity.api.tracker.SimpleDataTracker;
import it.unimi.dsi.fastutil.ints.IntList;
import org.jetbrains.annotations.Nullable;

import java.util.Optional;
import java.util.UUID;
import java.util.function.Consumer;
import net.minecraft.class_10182;
import net.minecraft.class_10264;
import net.minecraft.class_1297;
import net.minecraft.class_1299;
import net.minecraft.class_243;
import net.minecraft.class_2561;
import net.minecraft.class_2596;
import net.minecraft.class_2602;
import net.minecraft.class_2604;
import net.minecraft.class_2684;
import net.minecraft.class_2739;
import net.minecraft.class_3222;
import net.minecraft.class_3532;
import net.minecraft.class_4050;

@SuppressWarnings("ConstantConditions")
public abstract class GenericEntityElement extends AbstractElement {
    protected final DataTrackerLike dataTracker = this.createDataTracker();
    private final int id = VirtualEntityUtils.requestEntityId();
    private final UUID uuid = UUID.randomUUID();
    private float pitch;
    private float yaw;
    private boolean isRotationDirty;
    private boolean sendPositionUpdates = true;
    private boolean instantPositionUpdates = false;
    private boolean alwaysSyncAbsolutePosition = false;

    private int updatesSinceLastAbsolutePositionSync = 0;

    protected DataTrackerLike createDataTracker() {
        return new SimpleDataTracker(this.getEntityType());
    }

    public boolean isDirty() {
        return this.isRotationDirty || this.dataTracker.isDirty();
    }

    public boolean isRotationDirty() {
        return isRotationDirty;
    }

    public void ignorePositionUpdates() {
        setSendPositionUpdates(false);
    }
    public void instantPositionUpdates() {
        setInstantPositionUpdates(true);
    }

    public void setInstantPositionUpdates(boolean value) {
        this.instantPositionUpdates = value;
    }

    public void setSendPositionUpdates(boolean b) {
        this.sendPositionUpdates = b;
    }

    public boolean isSendingPositionUpdates() {
        return this.sendPositionUpdates;
    }

    public void setAlwaysSyncAbsolutePosition(boolean alwaysSyncAbsolutePosition) {
        this.alwaysSyncAbsolutePosition = alwaysSyncAbsolutePosition;
    }

    public boolean isAlwaysSyncingAbsolutePosition() {
        return alwaysSyncAbsolutePosition;
    }

    public void setPitch(float pitch) {
        if (this.pitch != pitch) {
            this.pitch = pitch;
            this.isRotationDirty = true;
        }
    }

    public void setYaw(float yaw) {
        if (this.yaw != yaw) {
            this.yaw = yaw;
            this.isRotationDirty = true;
        }
    }

    public float getYaw() {
        return this.yaw;
    }

    public float getPitch() {
        return this.pitch;
    }

    @Override
    public IntList getEntityIds() {
        return IntList.of(id);
    }

    public final UUID getUuid() {
        return this.uuid;
    }

    public final int getEntityId() {
        return this.id;
    }

    protected abstract class_1299<? extends class_1297> getEntityType();

    @Override
    public void startWatching(class_3222 player, Consumer<class_2596<class_2602>> packetConsumer) {
        packetConsumer.accept(this.createSpawnPacket(player));

        this.sendChangedTrackerEntries(player, packetConsumer);
    }

    protected class_2596<class_2602> createSpawnPacket(class_3222 player) {
        if (this.lastSyncedPos == null) {
            this.lastSyncedPos = this.getCurrentPos();
        }
        return new class_2604(this.id, this.uuid, this.lastSyncedPos.field_1352, this.lastSyncedPos.field_1351, this.lastSyncedPos.field_1350, this.pitch, this.yaw, this.getEntityType(), 0, class_243.field_1353, this.yaw);
    }

    protected void sendChangedTrackerEntries(class_3222 player, Consumer<class_2596<class_2602>> packetConsumer) {
        var changed = this.dataTracker.getChangedEntries();

        if (changed != null) {
            packetConsumer.accept(new class_2739(this.id, changed));
        }
    }

    @Override
    public void notifyMove(class_243 oldPos, class_243 newPos, class_243 delta) {
        if (this.sendPositionUpdates && this.instantPositionUpdates) {
            this.sendPositionUpdates();
        }
    }

    @Override
    public void stopWatching(class_3222 player, Consumer<class_2596<class_2602>> packetConsumer) {
    }

    @Override
    public void tick() {
        this.sendTrackerUpdates();
        if (this.sendPositionUpdates) {
            this.sendPositionUpdates();
        }
        this.sendRotationUpdates();
    }

    protected void sendPositionUpdates() {
        if (this.getHolder() == null) {
            return;
        }
        class_2596<class_2602> packet = null;
        var pos = this.getCurrentPos();

        if (pos.equals(this.lastSyncedPos)) {
            return;
        }

        if (this.lastSyncedPos == null || this.alwaysSyncAbsolutePosition || this.updatesSinceLastAbsolutePositionSync > 200) {
            packet = new class_10264(this.id, new class_10182(pos, class_243.field_1353, this.yaw, this.pitch), false);
            this.updatesSinceLastAbsolutePositionSync = 0;
        } else {
            packet = VirtualEntityUtils.createMovePacket(this.id, this.lastSyncedPos, pos, this.isRotationDirty, this.yaw, this.pitch);
            this.updatesSinceLastAbsolutePositionSync++;
        }

        if (packet != null) {
            this.getHolder().sendPacket(packet);
            if (!(packet instanceof class_2684.class_2687)) {
                this.lastSyncedPos = pos;
            }
        }
        this.isRotationDirty = false;
    }

    protected void sendTrackerUpdates() {
        if (this.dataTracker.isDirty()) {
            var dirty = this.dataTracker.getDirtyEntries();
            if (dirty != null) {
                this.getHolder().sendPacket(new class_2739(this.id, dirty));
            }
        }
    }

    protected void sendRotationUpdates() {
        if (this.isRotationDirty) {
            var i = class_3532.method_15375(yaw * 256.0F / 360.0F);
            var j = class_3532.method_15375(pitch * 256.0F / 360.0F);
            this.getHolder().sendPacket(new class_2684.class_2687(id, (byte) i, (byte) j, false));
            this.isRotationDirty = false;
        }
    }

    public DataTrackerLike getDataTracker() {
        return this.dataTracker;
    }

    public class_4050 getPose() {
        return this.dataTracker.get(EntityTrackedData.POSE);
    }

    public void setPose(class_4050 pose) {
        this.dataTracker.set(EntityTrackedData.POSE, pose);
    }

    public void setOnFire(boolean onFire) {
        this.setFlag(EntityTrackedData.ON_FIRE_FLAG_INDEX, onFire);
    }

    protected boolean getFlag(int index) {
        return (this.dataTracker.get(EntityTrackedData.FLAGS) & 1 << index) != 0;
    }

    protected void setFlag(int index, boolean value) {
        byte b = this.dataTracker.get(EntityTrackedData.FLAGS);
        if (value) {
            this.dataTracker.set(EntityTrackedData.FLAGS, (byte) (b | 1 << index));
        } else {
            this.dataTracker.set(EntityTrackedData.FLAGS, (byte) (b & ~(1 << index)));
        }

    }

    public boolean isSneaking() {
        return this.getFlag(EntityTrackedData.SNEAKING_FLAG_INDEX);
    }

    public void setSneaking(boolean sneaking) {
        this.setFlag(EntityTrackedData.SNEAKING_FLAG_INDEX, sneaking);
    }

    public boolean isSprinting() {
        return this.getFlag(EntityTrackedData.SPRINTING_FLAG_INDEX);
    }

    public void setSprinting(boolean sprinting) {
        this.setFlag(EntityTrackedData.SPRINTING_FLAG_INDEX, sprinting);
    }

    public boolean isGlowing() {
        return this.getFlag(EntityTrackedData.GLOWING_FLAG_INDEX);
    }

    public final void setGlowing(boolean glowing) {
        this.setFlag(EntityTrackedData.GLOWING_FLAG_INDEX, glowing);
    }

    public boolean isInvisible() {
        return this.getFlag(EntityTrackedData.INVISIBLE_FLAG_INDEX);
    }

    public void setInvisible(boolean invisible) {
        this.setFlag(EntityTrackedData.INVISIBLE_FLAG_INDEX, invisible);
    }

    public int getAir() {
        return this.dataTracker.get(EntityTrackedData.AIR);
    }

    public void setAir(int air) {
        this.dataTracker.set(EntityTrackedData.AIR, air);
    }

    public int getFrozenTicks() {
        return this.dataTracker.get(EntityTrackedData.FROZEN_TICKS);
    }

    public void setFrozenTicks(int frozenTicks) {
        this.dataTracker.set(EntityTrackedData.FROZEN_TICKS, frozenTicks);
    }

    @Nullable
    public class_2561 getCustomName() {
        return this.dataTracker.get(EntityTrackedData.CUSTOM_NAME).orElse(null);
    }

    public void setCustomName(@Nullable class_2561 name) {
        this.dataTracker.set(EntityTrackedData.CUSTOM_NAME, Optional.ofNullable(name));
    }

    public boolean isCustomNameVisible() {
        return Boolean.TRUE == this.dataTracker.get(EntityTrackedData.NAME_VISIBLE);
    }

    public void setCustomNameVisible(boolean visible) {
        this.dataTracker.set(EntityTrackedData.NAME_VISIBLE, visible);
    }

    public boolean isSilent() {
        return Boolean.TRUE == this.dataTracker.get(EntityTrackedData.SILENT);
    }

    public void setSilent(boolean silent) {
        this.dataTracker.set(EntityTrackedData.SILENT, silent);
    }

    public boolean hasNoGravity() {
        return Boolean.TRUE == this.dataTracker.get(EntityTrackedData.NO_GRAVITY);
    }

    public void setNoGravity(boolean noGravity) {
        this.dataTracker.set(EntityTrackedData.NO_GRAVITY, noGravity);
    }

    public void setRotation(float pitch, float yaw) {
        this.setPitch(pitch);
        this.setYaw(yaw);
    }
}
